/** @file
 * @brief PadicoTM components definition and API
 * @note shouldn't be included directly: #include <Padico/Puk.h> instead.
 * @ingroup Puk
 */

#ifndef TM_COMPONENTS_H
#define TM_COMPONENTS_H


/** @defgroup PadicoComponents API: PadicoTM components description
 * @ingroup Puk
 */

/** @defgroup PadicoComponentRegistry API: PadicoTM components registry- registration and query
 * @ingroup Puk
 */

/** @defgroup PadicoAssembly API: PadicoTM components assembly and instanciation interface
 * @see @ref PadicoComponents
 * @ingroup Puk
 */

/** @defgroup PadicoAdapter List of PadicoTM adapters
 */

/* ********************************************************* */
/* *** Components description ****************************** */


/** @addtogroup PadicoComponents
 * @{
 */

/* *** Types *********************************************** */

typedef struct puk_iface_s*puk_iface_t;
typedef struct puk_facet_s*puk_facet_t;
typedef struct puk_adapter_s*puk_adapter_t;

/* *** Facets and attributes ******************************* */

/** An interface from a component
 */
struct puk_iface_s
{
  const char*name;
};
PUK_VECT_TYPE(puk_iface, struct puk_iface_s*);

/** Attach a driver to an iface (e.g. for iface 'foo'):
 *    -- create an iface lookup function
 *    -- create the receptacle type
 *    -- typed context and instance indirections
 * Prototypes:
 *   padico_iface_t padico_iface_foo(void);
 *   struct puk_receptacle_foo_s;
 *   void puk_instance_indirect_foo(puk_instance_t, struct puk_receptacle_foo_s*);
 *   void puk_context_indirect_foo(puk_instance_t, puk_context_t, struct puk_receptacle_foo_s*);
 *   const struct foo_driver_s*puk_adapter_get_driver_foo(puk_adapter_t);
 */
#define PUK_IFACE_TYPE(IFACE_NAME, DRIVER_TYPE) \
  static inline puk_iface_t puk_iface_##IFACE_NAME(void) \
  { static puk_iface_t _iface = NULL; \
    if(!_iface) _iface = puk_iface_register(#IFACE_NAME); \
    return _iface; } \
  struct puk_receptacle_##IFACE_NAME##_s \
  { \
    DRIVER_TYPE*driver; \
    void*_status; \
  }; \
  static inline void puk_instance_indirect_##IFACE_NAME(puk_instance_t instance, \
                                                        struct puk_receptacle_##IFACE_NAME##_s*receptacle) \
  { puk_instance_indirect(instance, puk_iface_##IFACE_NAME(), &receptacle->driver, &receptacle->_status); } \
  static inline void puk_context_indirect_##IFACE_NAME(puk_instance_t instance, puk_context_t context, \
                                                       struct puk_receptacle_##IFACE_NAME##_s*receptacle) \
  { puk_context_indirect(instance, context, puk_iface_##IFACE_NAME(), &receptacle->driver, &receptacle->_status); } \
  static inline const DRIVER_TYPE*puk_adapter_get_driver_##IFACE_NAME(puk_adapter_t adapter) \
  { return (DRIVER_TYPE*) puk_adapter_get_driver(adapter, puk_iface_##IFACE_NAME()); }

/** A facet from a component
 */
struct puk_facet_s
{
  puk_iface_t iface; /**< interface exposed by the facet */
  const char*label;        /**< label (name) of the facet */
  const void*driver;       /**< driver that implements the facet */
};
PUK_VECT_TYPE(puk_facet, struct puk_facet_s);

/** A receptacle from a component
 */
struct puk_uses_s
{
  puk_iface_t iface;
  const char*label;
};
PUK_VECT_TYPE(puk_uses, struct puk_uses_s);

/* *** Adapter ********************************************* */

/** An abstract adapter descriptor 
 */
struct puk_adapter_s
{
  const char*name;     /**< entry name (eg. "Circuit/SysIO") )*/
  puk_hashtable_t attrs; /**< attributes table: label -> value @note for undef attr: value=&attr_undefined */
  struct puk_uses_vect_s  uses;     /**< the interfaces needed by this adapter */
  struct puk_facet_vect_s provides; /**< the interfaces provided by this adapter */
};
PUK_VECT_TYPE(puk_adapter, puk_adapter_t);

const void*puk_adapter_get_driver(puk_adapter_t a, puk_iface_t iface);

/** @} */

/* *** Register & query ************************************ */

/** @addtogroup PadicoComponentRegistry
 * @{
 */

/* ** interface registry */

puk_iface_t   puk_iface_register(const char*name);
puk_iface_t   puk_iface_lookup(const char*name);

/* ** adapter registry */

puk_adapter_t puk_adapter_resolve(const char*adapter_name);

/* ** registry locking handlers */
void padico_tm_comp_set_handlers(void(*_acquire)(void), void(*_release)(void));


/* *** Shortcuts for easy declaration ********************** */

struct puk_component_record_s
{
  enum { TM_ADAPTER_PROVIDES, TM_ADAPTER_USES, TM_ADAPTER_ATTR } kind;
  union
  {
    struct
    {
      const char*iface_name;
      const void*driver;
    } provides;
    struct
    {
      const char*iface_name;
    } uses;
    struct
    {
      const char*label;
      const char*value;
    } attr;
  } content;
};
#define puk_adapter_provides(IFACE_NAME, DRIVER) \
  ((struct puk_component_record_s){ \
    .kind = TM_ADAPTER_PROVIDES, \
    .content = {.provides = { .iface_name = (IFACE_NAME), .driver = (DRIVER)} }})
#define puk_adapter_uses(IFACE_NAME) \
  ((struct puk_component_record_s){ \
    .kind = TM_ADAPTER_USES, \
    .content = { .uses = { .iface_name = (IFACE_NAME) } } })
#define puk_adapter_attr(LABEL, VALUE) \
  ((struct puk_component_record_s){ \
    .kind = TM_ADAPTER_ATTR, \
    .content = { .attr = { .label = (LABEL), .value = (VALUE) } } })

puk_adapter_t 
puk_adapter_declare2(const char*adapter_name, int n_records,
			   const struct puk_component_record_s records[]);

#define puk_adapter_declare(ADAPTER_NAME, RECORDS...) \
  puk_adapter_declare2((ADAPTER_NAME), \
    sizeof((struct puk_component_record_s[]){ RECORDS }) / \
     sizeof(struct puk_component_record_s), \
    ((struct puk_component_record_s[]){ RECORDS }))


/** @} */

/* *** Adapter-specific data ******************************* */

/** @addtogroup PadicoComponents
 * @{
 */

struct puk_adapter_specific_bucket_s
{
  puk_adapter_t adapter;
  void*content;
};
PUK_VECT_TYPE(puk_adapter_specific_bucket, struct puk_adapter_specific_bucket_s);

typedef struct puk_adapter_specific_bucket_vect_s puk_adapter_specific_t;

#ifndef __cplusplus
static inline void puk_adapter_specific_init(puk_adapter_specific_t*as)
  __attribute__((deprecated));
static inline void puk_adapter_specific_destroy(puk_adapter_specific_t*as)
  __attribute__((deprecated));
static inline void*puk_adapter_getspecific(puk_adapter_specific_t*as,
						 puk_adapter_t a)
  __attribute__((deprecated));
static inline void puk_adapter_setspecific(puk_adapter_specific_t*as,
						 puk_adapter_t a,
						 void*value)
  __attribute__((deprecated));
     
static inline void puk_adapter_specific_init(puk_adapter_specific_t*as)
{
  puk_adapter_specific_bucket_vect_init(as);
}
static inline void puk_adapter_specific_destroy(puk_adapter_specific_t*as)
{
  puk_adapter_specific_bucket_vect_destroy(as);
}
static inline void*puk_adapter_getspecific(puk_adapter_specific_t*as,
						 puk_adapter_t a)
{
  puk_adapter_specific_bucket_vect_itor_t i;
  puk_vect_foreach(i, puk_adapter_specific_bucket, as)
    {
      if(i->adapter == a)
	return i->content;
    }
  return NULL;
}
static inline void puk_adapter_setspecific(puk_adapter_specific_t*as,
						 puk_adapter_t a,
						 void*value)
{
  puk_adapter_specific_bucket_vect_itor_t i;
  puk_vect_foreach(i, puk_adapter_specific_bucket, as)
    {
      if(i->adapter == a)
	{
	  i->content = value;
	  return;
	}
    }
  struct puk_adapter_specific_bucket_s b = { .adapter = a, .content = value };
  puk_adapter_specific_bucket_vect_push_back(as, b);
}
#endif /* __cplusplus */

/** @} */


/* ********************************************************* */
/* *** Assembly ******************************************** */

/** @addtogroup PadicoAssembly
 * @{
 */

/* *** Types ********************************************* */

typedef struct puk_component_conn_s*    puk_component_conn_t;
typedef struct puk_context_s* puk_context_t;
typedef struct puk_instance_s*puk_instance_t;

PUK_VECT_TYPE(puk_assembly, puk_adapter_t);

/* *** Adapter context *********************************** */

/** a connected receptacle */
struct puk_component_conn_s
{
  puk_iface_t     iface;
  puk_context_t context;
  const void*driver;
};
PUK_VECT_TYPE(puk_component_conn, struct puk_component_conn_s);


/** an adapter in its context (assembly, connections, and attributes) */
struct puk_context_s
{
  const char*         id; /**< context ID, as given in the XML assembly description */
  puk_adapter_t  adapter; /**< the adapter this context corresponds to */
  puk_adapter_t assembly; /**< the assembly this context belongs to */
  struct puk_component_conn_vect_s conns; /**< interfaces connections */
  puk_hashtable_t  attrs; /**< attributes */
};
PUK_VECT_TYPE(puk_context, puk_context_t);

const char*puk_context_getattr(puk_context_t, const char*);

/* *** Composite ******************************************* */

/** an assembly made of several adapters and their connections */
struct puk_composite_content_s
{
  struct puk_composite_driver_s
  {
    struct puk_composite_content_s*content;
  } driver;
  struct puk_component_conn_vect_s entry_points; /**< entry point of the assembly */
  puk_hashtable_t contexts; /**< contexts in this assembly- hashed by context ID */
  int checked; /**< whether the assembly as been checked (ok to be instanciated) */
};

padico_string_t      puk_adapter_serialize(puk_adapter_t);

puk_adapter_t puk_adapter_parse(const char*);


/* ********************************************************* */
/* *** Instanciation *************************************** */


/* *** Assembly instance ********************************* */

/** An instance of an adapter */
struct puk_instance_s
{
  puk_adapter_t adapter; /**< the adapter of which this is an instance */
  void*status;                 /**< adapter-specific status */
};

/** Status of a composite adapter */
struct puk_composite_status_s
{
  puk_hashtable_t status_table; /**< hash: context -> status */
  puk_instance_t instance;
};

/** a predefined descriptor for the instanciation interface: PadicoAdapter */
struct puk_adapter_driver_s
{
  void*(*instanciate)(puk_instance_t, puk_context_t);
  void (*destroy)(void*);
};

/** Creates a instance of a given assembly. */
puk_instance_t puk_adapter_instanciate(puk_adapter_t adapter);

/** Destroys an instance */
void puk_instance_destroy(puk_instance_t assembly);


/** Finds driver & status of a sub-context */
void puk_context_indirect(puk_instance_t instance,
				puk_context_t context,
				puk_iface_t iface,
				void*_driver, void *_status);

/** Finds an entry point of an instance */
void puk_instance_indirect(puk_instance_t instance,
				 puk_iface_t iface,
				 void*_driver, void*_status);


/* *** various inline functions */

PUK_IFACE_TYPE(PadicoAdapter, struct puk_adapter_driver_s);
PUK_IFACE_TYPE(PadicoComposite, struct puk_composite_driver_s);

#ifndef __cplusplus
/** Finds adapter instance status for a given (self) context */
static inline
void*puk_context_self_status(puk_instance_t instance,
				   puk_context_t context)
{
  if(context)
    {
      const struct puk_composite_status_s*composite_status = instance->status;
      return puk_hashtable_lookup(composite_status->status_table, context);
    }
  else
    return instance->status;
}
#endif /* __cplusplus */




#ifndef __cplusplus
static inline puk_component_conn_t puk_adapter_get_facet(puk_adapter_t, puk_iface_t) __attribute__((deprecated));
static inline
puk_component_conn_t puk_adapter_get_facet(puk_adapter_t assembly,
					      puk_iface_t iface)
{
  const struct puk_composite_driver_s*const driver =
    puk_adapter_get_driver(assembly, puk_iface_PadicoComposite());
  if(!driver)
    {
      padico_fatal("PadicoTM: adapter *%s* has no 'PadicoComposite' facet.\n", assembly->name);
    }
  int i;
  puk_component_conn_t entry_point = NULL;
  padico_warning("PadicoTM: obsolete puk_adapter_get_facet() called!\n");
  for(i = 0; i < puk_component_conn_vect_size(&driver->content->entry_points); i++)
    {
      if(puk_component_conn_vect_at(&driver->content->entry_points, i).iface == iface)
	entry_point = puk_component_conn_vect_ptr(&driver->content->entry_points, i);
    }
  return entry_point;
}
#endif /* __cplusplus */


/** @} */

/* *** Enumerator **************************************** */

/* Adapter enumerators provide an _ordered_ enumeration of 
 * the contexts of an assembly, so as to be able to calculate
 * inter-adapter dependancies when instanciating assemblies.
 */

typedef struct puk_adapter_enumerator_s*puk_adapter_enumerator_t;

/** direction of adapter enumeration : top-down, or bottom-up
 */
enum adapter_enumerator_way_e
  { 
    TOP_DOWN = 1,
    BOTTOM_UP
  };

puk_adapter_enumerator_t puk_adapter_enumerator_new(puk_adapter_t assembly, int way);
puk_context_t         puk_adapter_enumerator_next(puk_adapter_enumerator_t e);
void                        puk_adapter_enumerator_delete(puk_adapter_enumerator_t e);


/* ********************************************************* */


void puk_adapter_check(puk_adapter_t assembly);

#endif /* TM_COMPONENTS_H */

