/** @file
 * @note This file should be "clean", i.e. it must *not* include 
 * any other header.
 */

#include "puk_config.h"

/** whether PukABI should wrap libc internals (i.e. __libc_* functions) */
#undef  PUK_ABI_WRAP_INTERNALS
/** PukABI should try to recover incorrect mix of hfd/abi_fd (needed for libresolv) */
#define PUK_ABI_RECOVER_BADFD
/** PukABI should show warnings on bad fd automatic recovery */
#undef  PUK_ABI_SHOW_RECOVERY
/** activate PukABI traces */
#undef  PUK_ABI_TRACE_ON 
/** show pid in PukABI traces */
#undef  PUK_ABI_TRACE_PID
/** trace spinlocks done by PukABI */
#undef  PUK_ABI_TRACE_SPINLOCK
/** enable/disable assertion check */
#undef  PUK_ABI_CHECK_ASSERT
/** show stack top pointer in traces */
#undef  PUK_ABI_SHOW_STACKTOP

/** Maximum fd allocated by PukABI- @todo check getdtablesize(), RLIMIT_NOFILE, FD_SETSIZE or OPEN_MAX for a better value */
#define PUK_ABI_FD_MAX     1024
/** Offset for the fd virtualizer */
#define PUK_ABI_FD_OFFSET  32
/** End of stdio fd interval */
#define PUK_ABI_FD_STDIO   2


/* *** Puk ABI tools
 */
__PUK_SYM_INTERNAL void puk_abi_trace(char*msg, ...);
__PUK_SYM_INTERNAL void puk_abi_print(char*msg, ...);
__PUK_SYM_INTERNAL void puk_abi_fatal(char*msg, ...);
__PUK_SYM_INTERNAL const void*puk_sym_resolve_next(const char*symbol);
#ifdef PUK_ABI_CHECK_ASSERT
__PUK_SYM_INTERNAL void puk_abi_do_assert(int cond, const char*file, int line, const char*func);
#  define puk_abi_assert(COND) puk_abi_do_assert(COND, __FILE__, __LINE__, __FUNCTION__)
#else
#  define puk_abi_assert(COND)
#endif


/* *** Spinlock
 * Duplicated from Puk.h. TODO.
 */
typedef void (*puk_spinlock_handler_t)(void);
struct puk_spinlock_s
{
  void (*acquire)(void);
  void (*release)(void);
  int  (*trylock)(void); /* 1: lock busy; 0: lock free */
};
#ifdef PUK_ABI_TRACE_SPINLOCK
#define puk_spinlock_acquire() \
  (puk_spinlock.acquire?puk_abi_print("A: %s", __FUNCTION__),(*puk_spinlock.acquire)():(void)0)
#define puk_spinlock_release() \
  (puk_spinlock.release?(*puk_spinlock.release)(),puk_abi_print("R: %s", __FUNCTION__):(void)0)
#define puk_spinlock_trylock() \
  (puk_spinlock.trylock?puk_abi_print("T: %s", __FUNCTION__),(*puk_spinlock.trylock)():0)
#else
#define puk_spinlock_acquire() (*puk_spinlock.acquire)()
#define puk_spinlock_release() (*puk_spinlock.release)()
#define puk_spinlock_trylock() (*puk_spinlock.trylock)()
#endif

extern struct puk_spinlock_s puk_spinlock;

/* *** File descriptors
 */

/** dump table of known file descriptors */
__PUK_SYM_INTERNAL void puk_fds_dump(void);
/** checks whether the given fd is real or virtual */
__PUK_SYM_INTERNAL int  puk_fds_isreal(int abi_fd);
/** converts a file descriptor from the PukABI numbering into real/virtual
 * @note use puk_fds_isreal() to know whether the returned fd is real or virtual.
 */
__PUK_SYM_INTERNAL int  puk_fds_from_abi(int abi_fd);
/** converts a real fd number into a PukABI fd */
__PUK_SYM_INTERNAL int  puk_fds_lookup(int phys_fd, int kind);
__PUK_SYM_INTERNAL void puk_fdset_from_abi(void*fdset, int fdmax);
__PUK_SYM_INTERNAL void puk_fdset_to_abi(void*fdset, int fdmax, int isreal);
__PUK_SYM_INTERNAL int  puk_fdset_isreal(const void*fds1, const void*fds2, const void*fds3, int fdmax);
__PUK_SYM_INTERNAL void puk_pollfd_from_abi(void*ptr, int nfds);
__PUK_SYM_INTERNAL void puk_pollfd_to_abi(void*ptr, int nfds);

/* *** oracles for variadic functions
 */
__PUK_SYM_INTERNAL int puk_abi_oracle__fcntl(int cmd);
__PUK_SYM_INTERNAL int puk_abi_oracle__open(int oflag);

/* *** memory management
 */
__PUK_SYM_INTERNAL void puk_abi_mem_init(void);

/* *** Symbols
 */
struct puk_symbol_s
{
  const void*ptr_real;     /**< NULL means "not initialized" => lookup */
  const void*ptr_virtual;  /**< NULL means "not initialized" => use real  */
  const char name[];       /**< name of the symbol */
};

extern const void*puk_sym_resolve(struct puk_symbol_s*sym);

#define PUK_SYM_FUNCTYPE(SYM) _PASTE(SYM, _func_t)
#define PUK_SYM_DESCNAME(SYM) _PASTE3(__puk_abi_desc__, SYM, __)
#define PUK_SYM_STUB(SYM)     _PASTE3(__puk_abi_stub__, SYM, __)
#define PUK_SYM_WRAP(SYM)     _PASTE(__puk_abi_wrap__, SYM)

#define PUK_SYM_REAL(SYM)    ((PUK_SYM_DESCNAME(SYM).ptr_real)?:\
                               puk_sym_resolve(&PUK_SYM_DESCNAME(SYM)))

#define PUK_SYM_VIRTUAL(SYM) ((PUK_SYM_DESCNAME(SYM).ptr_virtual)?:\
                              (puk_abi_fatal("Puk-ABI: virtual symbol not found: *%s*\n##### Abort.\n",\
                                             _TOSTRING(SYM)),NULL))

/** This macro declares the types needed for a PukABI-managed symbol
 */
#define PUK_ABI_MKTYPE(SYM, RETTYPE, ARGSDECL) \
  __PUK_SYM_INTERNAL RETTYPE PUK_SYM_STUB(SYM) ARGSDECL; /* stub:   __puk_abi_stub__foo */ \
  extern RETTYPE PUK_SYM_WRAP(SYM) ARGSDECL;             /* wrap:   __puk_abi_wrap__foo */ \
  extern struct puk_symbol_s PUK_SYM_DESCNAME(SYM);      /* struct: __puk_abi_desc__foo__ */ \
  typedef RETTYPE (*PUK_SYM_FUNCTYPE(SYM)) ARGSDECL;     /* func type: foo_func_t */



/* *** errno
 * POSIX source in _REENTRANT mode requires 'errno' to be a macro
 * of the form: *((*func)())
 * We must guess the name of 'func', though :-)
 */

#if defined(PADICO_SYMBOL_ERRNO)
#  define puk_abi_errno_sym PADICO_SYMBOL_ERRNO
#else
#warning "Puk-ABI: PADICO_SYMBOL_ERRNO not defined in autoconf. Trying to guess."
#  if defined(linux)
#    define puk_abi_errno_sym __errno_location
#  elif defined(sun)
#    define puk_abi_errno_sym ___errno
#  else
#    error "Puk-ABI: unsupported OS. Cannot guess pointer to errno."
#  endif
#endif

extern int*puk_abi_errno_stub(void); /* actually, an alias to PUK_SYM_STUB(__errno_location) */

#define puk_abi_errno_real() \
 (*((PUK_SYM_FUNCTYPE(puk_abi_errno_sym)) PUK_SYM_REAL(puk_abi_errno_sym)))()

#define PUK_ABI_ERRNO_INHERIT \
 { *(puk_abi_errno_stub()) = *(puk_abi_errno_real()); }

