/** @file
 */

#include "Puk-internals.h"
#include "Module.h"
#include <stdlib.h>

PADICO_MODULE_HOOK(Puk);


/** PadicoTM components manager 
 */
static struct
{
  void (*comp_acquire)(void);
  void (*comp_release)(void);
  puk_hashtable_t     ifaces;  /**< table of known interfaces- hashed by name */
  puk_hashtable_t   adapters;  /**< table of registered adapters- hashed by name */
  const char attr_undefined[];
} tm_comp = { .attr_undefined = "undefined", .comp_acquire = NULL, .comp_release = NULL };

/* ********************************************************* */

static void tm_assembly_start_handler(puk_parse_entity_t e);
static void tm_assembly_end_handler(puk_parse_entity_t e);
static void tm_adapter_start_handler(puk_parse_entity_t e);
static void tm_adapter_end_handler(puk_parse_entity_t e);
static void tm_attr_end_handler(puk_parse_entity_t e);
static void tm_uses_start_handler(puk_parse_entity_t e);
static void tm_assembly_provides_start_handler(puk_parse_entity_t e);

/* ********************************************************* */
/* local shortcuts */

/** get a receptacle of an adapter context */
static inline
puk_component_conn_t tm_context_getused(puk_context_t context,
				    puk_iface_t iface)
{
  puk_component_conn_itor_t i;
  puk_vect_foreach(i, puk_component_conn, &context->conns)
    {
      if(i->iface == iface)
	return i;
    }
  return NULL;
}

/** @internal Gets the composite content from an adapter (supposed to be a composite) */
static inline
struct puk_composite_content_s*tm_composite_get(puk_adapter_t adapter)
{
  const struct puk_composite_driver_s*const driver = 
    puk_adapter_get_driver(adapter, puk_iface_PadicoComposite());
  if(!driver)
    {
      padico_fatal("PadicoTM: adapter *%s* has no 'PadicoComposite' facet.\n", adapter->name);
    }
  return driver->content;
}

void padico_tm_comp_set_handlers(void(*_acquire)(void), void(*_release)(void))
{
  tm_comp.comp_acquire = _acquire;
  tm_comp.comp_release = _release;
}

static inline void tm_comp_acquire(void)
{
  if(tm_comp.comp_acquire)
    (*tm_comp.comp_acquire)();
}
static inline void tm_comp_release(void)
{
  if(tm_comp.comp_release)
    (*tm_comp.comp_release)();
}

/* ********************************************************* */

void padico_puk_components_init(void)
{
  tm_comp.ifaces   = puk_hashtable_new(&puk_hash_string, &puk_hash_string_eq);
  tm_comp.adapters = puk_hashtable_new(&puk_hash_string, &puk_hash_string_eq);
  puk_iface_register("PadicoAdapter");
  puk_iface_register("PadicoComposite");

  /* parsing */
  puk_xml_add_action((struct puk_tag_action_s)
  {
    .xml_tag        = "NS:assembly",
    .start_handler  = &tm_assembly_start_handler,
    .end_handler    = &tm_assembly_end_handler,
    .required_level = PUK_TRUST_CONTROL
  });
  puk_xml_add_action((struct puk_tag_action_s)
  {
    .xml_tag        = "NS:adapter",
    .start_handler  = &tm_adapter_start_handler,
    .end_handler    = &tm_adapter_end_handler,
    .required_level = PUK_TRUST_CONTROL
  });
  puk_xml_add_action((struct puk_tag_action_s)
  {
    .xml_tag        = "NS:attr",
    .start_handler  = NULL,
    .end_handler    = &tm_attr_end_handler,
    .required_level = PUK_TRUST_CONTROL
  });
  puk_xml_add_action((struct puk_tag_action_s)
  {
    .xml_tag        = "NS:uses",
    .start_handler  = &tm_uses_start_handler,
    .end_handler    = NULL,
    .required_level = PUK_TRUST_CONTROL
  });
  puk_xml_add_action((struct puk_tag_action_s)
  {
    .xml_tag        = "NS:assembly-provides",
    .start_handler  = &tm_assembly_provides_start_handler,
    .end_handler    = NULL,
    .required_level = PUK_TRUST_CONTROL
  });

}

/* ********************************************************* */
/* *** PadicoComponent interface *************************** */

puk_iface_t puk_iface_register(const char*name)
{
  tm_comp_acquire();
  puk_iface_t iface = puk_hashtable_lookup(tm_comp.ifaces, name);
  if(!iface)
    {
      iface = padico_malloc(sizeof(struct puk_iface_s));
      padico_out(40, "registering iface %s\n", name);
      iface->name = padico_strdup(name);
      puk_hashtable_insert(tm_comp.ifaces, iface->name, iface);
    }
  tm_comp_release();
  return iface;
}

puk_iface_t puk_iface_lookup(const char*name)
{
  tm_comp_acquire();
  puk_iface_t iface = puk_hashtable_lookup(tm_comp.ifaces, name);
  if(!iface)
    {
      padico_out(6, "PadicoComponent: interface %s not found.\n", name);
    }
  tm_comp_release();
  return iface;
}

/* ********************************************************* */
/* ***** PadicoComponent adapter *************************** */

puk_adapter_t 
puk_adapter_declare2(const char*adapter_name, int n_records,
			   const struct puk_component_record_s records[])
{
  int i;
  puk_adapter_t a = padico_malloc(sizeof(struct puk_adapter_s));
  padico_out(40, "building adapter %s\n", adapter_name);
  a->name = padico_strdup(adapter_name);
  a->attrs = NULL;
  puk_uses_vect_init(&a->uses);
  puk_facet_vect_init(&a->provides);
  for(i = 0; i < n_records; i++)
    {
      struct puk_component_record_s r = records[i];
      switch(r.kind)
	{
	case TM_ADAPTER_PROVIDES:
	  {
	    puk_iface_t iface = puk_iface_lookup(r.content.provides.iface_name);
	    if(iface == NULL)
	      {
		padico_out(10, "auto-register interface %s (adapter: %s).\n",
			   r.content.provides.iface_name, adapter_name);
		iface = puk_iface_register(r.content.provides.iface_name);
	      }
	    padico_out(40, "  adapter provides facet %s\n", iface->name);
	    const struct puk_facet_s _facet =
	      {
		.iface  = iface,
		.label  = iface->name,
		.driver = r.content.provides.driver
	      };
	    puk_facet_vect_push_back(&a->provides, _facet);
	  }
 	  break;
	case TM_ADAPTER_USES:
	  {
	    puk_iface_t iface = puk_iface_lookup(r.content.uses.iface_name);
	    if(iface == NULL)
	      {
		padico_fatal("PadicoComponent: used interface %s not found- cannot register adapter %s.\n",
			     r.content.uses.iface_name, adapter_name);
	      }
	    padico_out(40, "  adapter uses iface %s\n", iface->name);
	    const struct puk_uses_s _uses = { .iface = iface, .label = iface->name };
	    puk_uses_vect_push_back(&a->uses, _uses);
	  }
	  break;
	case TM_ADAPTER_ATTR:
	  {
	    const char*label = r.content.attr.label;
	    const char*value = r.content.attr.value?:tm_comp.attr_undefined;
	    if(!a->attrs)
	      a->attrs = puk_hashtable_new_string();
	    puk_hashtable_insert(a->attrs, label, (void*)value);
	  }
	  break;
	}
    }
  padico_out(40, "registering adapter %s.\n", a->name);
  tm_comp_acquire();
  puk_hashtable_insert(tm_comp.adapters, a->name, a);
  tm_comp_release();
  return a;
}

const void*puk_adapter_get_driver(puk_adapter_t a, puk_iface_t iface)
{
  puk_facet_itor_t i;
  /* @note linear search better than anything else when #facet < 5! */
  puk_vect_foreach(i, puk_facet, &a->provides)
    {
      if(i->iface == iface)
	return i->driver;
    }
  return NULL;
}

/** Adapter lookup+auto-load if lookup fails.
 */
puk_adapter_t puk_adapter_resolve(const char*adapter_name)
{
  tm_comp_acquire();
  puk_adapter_t adapter = puk_hashtable_lookup(tm_comp.adapters, adapter_name);
  tm_comp_release();
  if(!adapter)
    {
      padico_mod_t mod = NULL;
      padico_rc_t rc = padico_puk_mod_resolve(&mod, adapter_name);
      padico_out(10, "adapter %s not found. Auto-load...\n", adapter_name);
      if(!padico_rc_iserror(rc))
	{
	  padico_out(6, "loading adapter %s ...\n", adapter_name);
	  puk_mod_set_ancestor(mod, padico_module_self());
	  rc = padico_puk_mod_load(mod);
	  tm_comp_acquire();
	  adapter = puk_hashtable_lookup(tm_comp.adapters, adapter_name);
	  tm_comp_release();
	}
      if(!padico_rc_iserror(rc))
	{
	  padico_out(10, "adapter %s successfully loaded.\n", adapter_name);
	}
      else
	{
	  padico_warning("PadicoTM: adapter %s auto-load failed.\n", adapter_name);
	}
    }
  return adapter;
}


/* ********************************************************* */
/* *** Adapter-specific ************************************ */

/* empty */


/* ********************************************************* */
/* *** Adapter context ************************************* */


void puk_context_putattr(puk_context_t context, 
			       const char*label, const char*value)
{
  if(!context->attrs)
    context->attrs = puk_hashtable_new_string();
  puk_hashtable_insert(context->attrs, padico_strdup(label), padico_strdup(value));
}

const char*puk_context_getattr(puk_context_t context,
				     const char*label)
{
  const char*value = context->attrs?puk_hashtable_lookup(context->attrs, label):NULL;
  return value;
}

void puk_context_indirect(puk_instance_t instance,
				puk_context_t context,
				puk_iface_t iface,
				void*_driver, void *_status)
{
  const puk_component_conn_t entry_point = tm_context_getused(context, iface);
  assert(entry_point != NULL);
  if(_driver)
    {
      *(const void**)_driver = entry_point->driver;
    }
  if(_status)
    {
      const struct puk_composite_status_s*composite_status = instance->status;
      *((void**)_status) = puk_hashtable_lookup(composite_status->status_table, entry_point->context);
    }
}


/* ********************************************************* */
/* *** Composite ******************************************* */

static void*tm_composite_instanciate(puk_instance_t instance, puk_context_t in_context)
{
  struct puk_composite_status_s*status = padico_malloc(sizeof(struct puk_composite_status_s));
  instance->status = status;
  if(in_context)
    {
      padico_fatal("PadicoTM: instances inside an instance not supported yet.\n"); /* TODO */
    }
  status->status_table = puk_hashtable_new_ptr();
  status->instance = instance;
  puk_context_t context = NULL;
  puk_adapter_enumerator_t e = puk_adapter_enumerator_new(instance->adapter, BOTTOM_UP);
  for(context  = puk_adapter_enumerator_next(e);
      context != NULL;
      context  = puk_adapter_enumerator_next(e))
    {
      const struct puk_adapter_driver_s*driver =
	puk_adapter_get_driver(context->adapter, puk_iface_PadicoAdapter());
      if(driver && driver->instanciate)
	{
	  void*sub_status = (*driver->instanciate)(instance, context);
	  puk_hashtable_insert(status->status_table, context, sub_status);
	}
      else
	{
	  padico_warning("PadicoTM: instanciating adapter without 'PadicoAdapter' iface.\n");
	}
    }
  puk_adapter_enumerator_delete(e);
  return status;
}

static void tm_composite_destroy(void*_status)
{
  struct puk_composite_status_s*composite_status = _status;
  puk_instance_t instance = composite_status->instance;
  puk_adapter_enumerator_t e = puk_adapter_enumerator_new(instance->adapter, BOTTOM_UP);
  puk_context_t context = NULL;
  for(context  = puk_adapter_enumerator_next(e);
      context != NULL;
      context  = puk_adapter_enumerator_next(e))
    {
      const struct puk_adapter_driver_s*driver =
	puk_adapter_get_driver(context->adapter, puk_iface_PadicoAdapter());
      if(driver && driver->destroy)
	{
	  void*sub_status = puk_context_self_status(instance, context);
	  (*driver->destroy)(sub_status);
	}
    }
  puk_hashtable_delete(composite_status->status_table);
  puk_adapter_enumerator_delete(e);
  padico_free(composite_status);
}

static struct puk_adapter_driver_s composite_adapter_driver =
  {
    .instanciate = &tm_composite_instanciate,
    .destroy     = &tm_composite_destroy
  };


/* ********************************************************* */
/* *** Adapter instance ************************************ */

puk_instance_t puk_adapter_instanciate(puk_adapter_t adapter)
{
  const struct puk_adapter_driver_s*driver =
    puk_adapter_get_driver(adapter, puk_iface_PadicoAdapter());
  puk_instance_t instance = padico_malloc(sizeof(struct puk_instance_s));
  instance->adapter = adapter;
  instance->status = NULL;
  if(driver && driver->instanciate)
    {
      instance->status = (*driver->instanciate)(instance, NULL);
    }
  return instance;
}

void puk_instance_destroy(puk_instance_t instance)
{
  const struct puk_adapter_driver_s*driver =
    puk_adapter_get_driver(instance->adapter, puk_iface_PadicoAdapter());
  if(driver && driver->destroy)
    {
      (*driver->destroy)(instance->status);
    }
  padico_free(instance);
}

void puk_instance_indirect(puk_instance_t instance,
				 puk_iface_t iface,
				 void*_driver, void*_status)
{
  const puk_adapter_t adapter = instance->adapter;
  if(_driver)
    {
      *((const void**)_driver) = puk_adapter_get_driver(adapter, iface);
    }
  if(_status)
    {
      const struct puk_composite_driver_s*composite_driver =
	puk_adapter_get_driver(adapter, puk_iface_PadicoComposite());
      if(composite_driver)
	{
	  struct puk_composite_content_s*content = composite_driver->content;
	  int i;
	  puk_component_conn_t entry_point = NULL;
	  for(i = 0; i < puk_component_conn_vect_size(&content->entry_points); i++)
	    {
	      if(puk_component_conn_vect_at(&content->entry_points, i).iface == iface)
		{
		  entry_point = puk_component_conn_vect_ptr(&content->entry_points, i);
		  break;
		}
	    }
	  assert(entry_point != NULL);
	  const struct puk_composite_status_s*composite_status = instance->status;
	  *((void**)_status) = puk_hashtable_lookup(composite_status->status_table, entry_point->context);
	}
      else
	{
	  *((void**)_status) = instance->status;
	}
    }
}


/* ********************************************************* */
/* *** Assembly marshalling ******************************** */

padico_string_t puk_adapter_serialize(puk_adapter_t assembly)
{
  padico_string_t s = padico_string_new();
  if(puk_adapter_get_driver(assembly, puk_iface_PadicoComposite()))
    {
      padico_string_catf(s, "<NS:assembly id=\"_puk:_x(%p):%s\">\n", assembly, assembly->name);
      /* adapters */
      puk_adapter_enumerator_t adapters = puk_adapter_enumerator_new(assembly, BOTTOM_UP);
      puk_context_t context = puk_adapter_enumerator_next(adapters);
      while(context)
	{
	  int i;
	  padico_string_catf(s, "  <NS:adapter name=\"%s\" id=\"%p\">\n", 
			     context->adapter->name, context);
	  /* receptacles (uses) */
	  for(i = 0; i < puk_component_conn_vect_size(&context->conns); i++)
	    {
	      puk_component_conn_t u = puk_component_conn_vect_ptr(&context->conns, i);
	      if(u->context != NULL)
		{
		  padico_string_catf(s, "    <NS:uses adapter-id=\"%p\" iface=\"%s\"/>\n", 
				     u->context, u->iface->name);
		}
	    }
	  /* attributes */
	  if(context->attrs)
	    {
	      puk_hashtable_enumerator_t attrs = puk_hashtable_enumerator_new(context->attrs);
	      char*label, *value; 
	      puk_hashtable_enumerator_next2(attrs, &label, &value);
	      while(label)
		{
		  padico_string_catf(s, "    <NS:attr label=\"%s\">%s</NS:attr>\n", 
				     label, value);
		  puk_hashtable_enumerator_next2(attrs, &label, &value);
		}
	      puk_hashtable_enumerator_delete(attrs);
	    }
	  padico_string_catf(s, "  </NS:adapter>\n");
	  context = puk_adapter_enumerator_next(adapters);
	}
      puk_adapter_enumerator_delete(adapters);
      /* exported facets */
      struct puk_composite_content_s*content = tm_composite_get(assembly);
      puk_component_conn_vect_itor_t e;
      puk_vect_foreach(e, puk_component_conn, &content->entry_points)
	{
	  padico_string_catf(s, "<NS:assembly-provides adapter-id=\"%p\" iface=\"%s\"/>\n",
			     e->context, e->iface->name);
	}
      padico_string_catf(s, "</NS:assembly>\n");
    }
  else
    {
      padico_string_catf(s, "  <NS:adapter name=\"%s\"/>\n", assembly->name);
    }
  return s;
}

puk_adapter_t puk_adapter_parse(const char*s)
{
  puk_adapter_t assembly = NULL;
  struct puk_parse_entity_s e = puk_xml_parse_buffer(s, strlen(s), PUK_TRUST_LOCAL);
  if(puk_parse_is(&e, "NS:assembly"))
    {
      assembly = puk_parse_get_content(&e);
    }
  else
    {
      padico_fatal("PadicoTM: error while parsing *%s*: not an assembly.\n", s);
    }
  return assembly;
}


/* ********************************************************* */
/* *** Assembly parsing ************************************ */


/* *** <NS:assembly> */
static void tm_assembly_start_handler(puk_parse_entity_t e)
{
  const char*id  = puk_parse_getattr(e, "id");
  const char*ref = puk_parse_getattr(e, "ref");
  puk_adapter_t assembly = NULL;
  assert(!ref != !id);
  if(ref)
    {
      tm_comp_acquire();
      assembly = puk_hashtable_lookup(tm_comp.adapters, ref);
      tm_comp_release();
    }
  else
    {
      assembly = padico_malloc(sizeof(struct puk_adapter_s));
      assembly->name = padico_strdup(id);
      assembly->attrs = NULL;
      puk_uses_vect_init(&assembly->uses);
      puk_facet_vect_init(&assembly->provides);
      const struct puk_facet_s adapter_facet =
	{
	  .iface  = puk_iface_PadicoAdapter(),
	  .driver = &composite_adapter_driver
	};
      puk_facet_vect_push_back(&assembly->provides, adapter_facet);
      struct puk_composite_content_s*composite_content = padico_malloc(sizeof(struct puk_composite_content_s));
      composite_content->driver.content = composite_content;
      const struct puk_facet_s composite_facet =
	{
	  .iface  = puk_iface_PadicoComposite(),
	  .driver = &composite_content->driver
	};
      puk_facet_vect_push_back(&assembly->provides, composite_facet);
      puk_component_conn_vect_init(&composite_content->entry_points);
      composite_content->contexts = puk_hashtable_new_string();
      composite_content->checked = 0;
      padico_out(40, "new assembly- id=%s\n", id);
    }
  puk_parse_set_content(e, assembly);
}
static void tm_assembly_end_handler(puk_parse_entity_t e)
{
  puk_adapter_t assembly = puk_parse_get_content(e);
  if(assembly)
    {
      struct puk_composite_content_s*content = tm_composite_get(assembly);
      if(puk_hashtable_size(content->contexts) == 0 ||
	 puk_component_conn_vect_size(&content->entry_points) == 0)
	{
	  padico_warning("PadicoTM: invalid assembly '%s'- skipping.\n", assembly->name);
	  padico_free(assembly); /* TODO: tm_adapter_destroy */
	  assembly = NULL;
	}
      else if(!puk_parse_getattr(e, "ref"))
	{
	  tm_comp_acquire();
	  const puk_adapter_t a = puk_hashtable_lookup(tm_comp.adapters, assembly->name);
	  if(a)
	    {
	      padico_out(6, "duplicate assembly '%s'\n", assembly->name);
	      assembly = a;
	      /* TODO: free assembly- memory leak here */
	    }
	  else
	    {
	      puk_hashtable_insert(tm_comp.adapters, assembly->name, assembly);
	    }
	  tm_comp_release();
	}
    }
  puk_parse_set_content(e, assembly);
}

/* *** <NS:adapter> */
static void tm_adapter_start_handler(puk_parse_entity_t e)
{
  const char*adapter_name = puk_parse_getattr(e, "name");
  const char*adapter_id   = puk_parse_getattr(e, "id");
  puk_adapter_t adapter = puk_adapter_resolve(adapter_name);
  if(adapter == NULL)
    {
      padico_warning("NetSelector: adapter '%s' not found- skipping.\n", adapter_name);
      puk_parse_set_content(e, NULL);
    }
  else
    {
      puk_adapter_t assembly = puk_parse_get_content(puk_parse_parent(e));
      puk_context_t context = padico_malloc(sizeof(struct puk_context_s));
      assert(adapter != NULL);
      context->adapter  = adapter;
      context->assembly = assembly;
      context->attrs    = NULL;
      /* instanciate receptacles */
      puk_component_conn_vect_init(&context->conns);
      {
	puk_uses_vect_itor_t i;
	puk_vect_foreach(i, puk_uses, &adapter->uses)
	  {
	    struct puk_component_conn_s uses =
	      { .iface = i->iface, .context = NULL, .driver = NULL };
	    puk_component_conn_vect_push_back(&context->conns, uses);
	  }
      }
      if(!adapter_id)
	{
	  padico_string_t sid = padico_string_new();
	  padico_string_printf(sid, "_puk:context-%s-(%p)", adapter->name, adapter);
	  context->id = padico_strdup(padico_string_get(sid));
	  padico_string_delete(sid);
	}
      else
	context->id = padico_strdup(adapter_id);
      puk_hashtable_insert(tm_composite_get(assembly)->contexts, context->id, context);


      padico_out(40, "created adapter name=%s; id=%s\n", adapter_name, adapter_id);
      puk_parse_set_content(e, context);
    }
}
static void tm_adapter_end_handler(puk_parse_entity_t e)
{
  puk_context_t context = puk_parse_get_content(e);
  if(context)
    {
      /* check receptacles */
      puk_component_conn_itor_t u;
      padico_rc_t rc = padico_rc_ok();
      puk_vect_foreach(u, puk_component_conn, &context->conns)
	{
	  if(!u->driver)
	    {
	      rc = padico_rc_error("Receptacle '%s' not connected in context '%s'.\n", 
				   u->iface->name, context->adapter->name);
	      padico_warning("PadicoTM: receptacle '%s' not connected in context '%s' (%s).\n", 
			     u->iface->name, context->id, context->adapter->name);
	    }
	}
      /* check attributes */
      if(context->adapter->attrs)
	{
	  puk_hashtable_enumerator_t e = puk_hashtable_enumerator_new(context->adapter->attrs);
	  const char*label = NULL;
	  char*value = NULL;
	  puk_hashtable_enumerator_next2(e, &label, &value);
	  while(label)
	    {
	      if((!context->attrs) || (context->attrs && !puk_hashtable_lookup(context->attrs, label)))
		{
		  if(value != tm_comp.attr_undefined)
		    {
		      /* take the default value */
		      puk_context_putattr(context, label, value);
		    }
		  else
		    {
		      rc = padico_rc_error("Attribute '%s' has no value in context '%s'.\n",
					   label, context->adapter->name);
		      padico_warning("PadicoTM: attribute '%s' has no value in context '%s'.\n",
				     label, context->adapter->name);
		      break;
		    }
		}
	      puk_hashtable_enumerator_next2(e, &label, &value);
	    }
	}
      puk_parse_set_rc(e, rc);
    }
}

/* *** <NS:attr> */
static void tm_attr_end_handler(puk_parse_entity_t e)
{
  const char*label = puk_parse_getattr(e, "label");
  const char*value = puk_parse_get_text(e);
  puk_context_t context = 
    puk_parse_get_content_checked(puk_parse_parent(e), "NS:adapter");
  if(context)
    {
      assert(context && label && value);
      puk_context_putattr(context, label, value);
      padico_out(40, "created attr %s\n", label);
    }
}

/* *** <NS:uses> */
static void tm_uses_start_handler(puk_parse_entity_t e)
{
  const char*adapter_id = puk_parse_getattr(e, "adapter-id");
  const char*iface_name = puk_parse_getattr(e, "iface");
  puk_iface_t iface = puk_iface_lookup(iface_name);
  if(!iface)
    padico_fatal("PadicoTM: invalid 'uses'- interfaces '%s' does not exist.\n", iface_name);
  puk_context_t user = puk_parse_get_content(puk_parse_parent(e));
  if(user)
    {
      const puk_adapter_t assembly = puk_parse_get_content(puk_parse_parent(puk_parse_parent(e)));
      if(!assembly)
	padico_fatal("PadicoTM: invalid 'uses'- not inside an assembly.\n");
      struct puk_composite_content_s*content = tm_composite_get(assembly);
      puk_context_t provider = puk_hashtable_lookup(content->contexts, adapter_id);
      if(provider)
	{
	  const void*const driver = puk_adapter_get_driver(provider->adapter, iface);
	  assert(driver != NULL);
	  puk_component_conn_t receptacle = tm_context_getused(user, iface);
	  assert(receptacle != NULL);
	  receptacle->iface   = iface;
	  receptacle->context = provider;
	  receptacle->driver  = driver;
	}
      else
	{
	  padico_warning("PadicoTM: invalid 'uses'- context '%s' does not exist.\n", adapter_id);
	}
    }
}

/* *** <NS:assembly-provides> */
static void tm_assembly_provides_start_handler(puk_parse_entity_t e)
{
  const char*adapter_id = puk_parse_getattr(e, "adapter-id");
  const char*iface_name = puk_parse_getattr(e, "iface");
  puk_iface_t iface = puk_iface_lookup(iface_name);
  if(!iface)
    padico_fatal("PadicoTM: interfaces '%s' does not exist.\n", iface_name);
  puk_adapter_t assembly = puk_parse_get_content(puk_parse_parent(e));
  struct puk_composite_content_s*content = tm_composite_get(assembly);
  if(assembly)
    {
      puk_context_t context = puk_hashtable_lookup(content->contexts, adapter_id);
      if(context)
	{
	  const void*driver = puk_adapter_get_driver(context->adapter, iface);
	  assert(iface != NULL && driver != NULL && context != NULL);
	  const struct puk_component_conn_s entry_point =
	    {
	      .iface   = iface,
	      .context = context,
	      .driver  = driver
	    };
	  puk_component_conn_vect_push_back(&content->entry_points, entry_point);
	  const struct puk_facet_s facet =
	    {
	      .iface = iface,
	      .driver = driver
	    };
	  puk_facet_vect_push_back(&assembly->provides, facet);
	}
    }
}


/* ********************************************************* */
/* *** Enumerator ****************************************** */

struct puk_adapter_enumerator_s
{
  puk_context_vect_t adapters;
  puk_context_itor_t current;
  enum adapter_enumerator_way_e way;
};

puk_context_t puk_adapter_enumerator_next(puk_adapter_enumerator_t e)
{
  puk_context_t a = NULL;
  switch(e->way)
    {
    case BOTTOM_UP:
      if(e->current != puk_context_vect_end(e->adapters))
	{
	  a = *e->current;
	  e->current = puk_context_vect_next(e->current);
	}
      else
	{
	  e->current = NULL;
	  a = NULL;
	}
      break;
    case TOP_DOWN:
      if(e->current != puk_context_vect_rend(e->adapters))
	{
	  a = *e->current;
	  e->current = puk_context_vect_rnext(e->current);
	}
      else
	{
	  e->current = NULL;
	  a = NULL;
	}
      break;
    }
  return a;
}

void puk_adapter_enumerator_delete(puk_adapter_enumerator_t e)
{
  puk_context_vect_delete(e->adapters);
  padico_free(e);
}

puk_adapter_enumerator_t
puk_adapter_enumerator_new(puk_adapter_t assembly, int way)
{
  puk_adapter_enumerator_t e = padico_malloc(sizeof(struct puk_adapter_enumerator_s));
  e->way      = way;
  e->adapters = puk_context_vect_new();
  e->current  = NULL;
  /* convert hashtable into unsorted vector
   */
  struct puk_composite_content_s*content = tm_composite_get(assembly);
  puk_hashtable_enumerator_t h = puk_hashtable_enumerator_new(content->contexts);
  puk_context_vect_t unsorted = puk_context_vect_new();
  puk_context_t c = puk_hashtable_enumerator_next_data(h);
  while(c)
    {
      assert(c->assembly == assembly);
      padico_out(60, "adding context id=%s; adapter=%s; assembly=%s\n", c->id, c->adapter->name, c->assembly->name);
      puk_context_vect_push_back(unsorted, c);
      c = puk_hashtable_enumerator_next_data(h);
    }
  puk_hashtable_enumerator_delete(h);
  /* topological sorting based on zero-in method (strictly speaking,
   * this is implemented as zero-out rather than zero-in) */
  const int degree = puk_context_vect_size(unsorted);
  int*matrix = padico_calloc(degree*degree, sizeof(int));
  int*done   = padico_calloc(degree, sizeof(int));
  /* matrix[x*degree + y] == 1 ~~ "x uses y" */
  int i, j, k;
  puk_context_t a_i, a_j;
  assert(degree > 0);
  /* convert the assembly into an adjacency matrix */
  for(i = 0; i < degree; i++)
    {
      a_i = puk_context_vect_at(unsorted, i);
      for(k = 0; k < puk_component_conn_vect_size(&a_i->conns); k++)
	{
	  for(j = 0; j < degree; j++)
	    {
	      a_j = puk_context_vect_at(unsorted, i);
	      if(puk_component_conn_vect_ptr(&a_i->conns, k)->context == a_j)
		{
		  matrix[i*degree+j] = 1;
		}
	    }
	}
    }
  for(k = 0; k < degree; k++)
    {
      for(i = 0; i < degree; i++)
	{
	  /* find a "zero-out" adapter- trying context 'i' */
	  padico_out(50, "searching for zero-out- k=%d; i=%d\n", k, i);
	  if(done[i])
	    {
	      padico_out(50, "i=%d is already done- skipping.\n", i);
	      goto next_i;
	    }
	  for(j = 0; j < degree; j++)
	    {
	      if(matrix[i*degree+j])
		{
		  padico_out(50, "i=%d is not zero-out- has out for j=%d\n", i, j);
		  goto next_i;
		}
	    }
	  padico_out(50, "i=%d is zero-out- clearing dependancies\n", i);
	  done[i] = 1;
	  /* clear all dependancies to context 'i' */
	  puk_context_t context = puk_context_vect_at(unsorted, i);
	  puk_context_vect_push_back(e->adapters, context);
	  for(j = 0; j < degree; j++)
	    {
	      matrix[j*degree+i] = 0;
	    }		  
	  goto next_k;
	next_i:
	  continue;
	}
    next_k:
      continue;
    }
  padico_free(matrix);
  padico_free(done);
  puk_context_vect_delete(unsorted);
  switch(way)
    {
    case BOTTOM_UP:
      e->current = puk_context_vect_begin(e->adapters);
      break;
    case TOP_DOWN:
      e->current = puk_context_vect_rbegin(e->adapters);
      break;
    }
  return e;
}


static
void puk_adapter_check_one(puk_adapter_t assembly)
{
  const struct puk_composite_driver_s*const driver = 
    puk_adapter_get_driver(assembly, puk_iface_PadicoComposite());
  if(driver)
    {  
      puk_adapter_enumerator_t e = puk_adapter_enumerator_new(assembly, TOP_DOWN);
      puk_adapter_enumerator_delete(e);
    }
}

void puk_adapter_check(puk_adapter_t assembly)
{
  if(assembly)
    {
      const struct puk_composite_driver_s*const driver =
	puk_adapter_get_driver(assembly, puk_iface_PadicoComposite());
      padico_print("PadicoTM: TMP- check assembly %p '%s' driver=%p ...\n", assembly, assembly->name, driver);
      puk_adapter_check_one(assembly);
      padico_print("PadicoTM: TMP- check assembly %p '%s'- done.\n", assembly, assembly->name);
    }
  puk_hashtable_enumerator_t h = puk_hashtable_enumerator_new(tm_comp.adapters);
  puk_adapter_t a = puk_hashtable_enumerator_next_data(h);
  while(a)
    {
      puk_adapter_check_one(a);
      a = puk_hashtable_enumerator_next_data(h);
    }
}
