/** @file
 * @brief Puk hashtables
 * @ingroup Puk
 */

#include "Puk-hashtable.h"

#undef DEBUG_HASH

#define puk_hashtable_init_size 11

static void puk_hashtable_rehash(puk_hashtable_t h);

static inline uint32_t puk_hash_to_index(uint32_t z, uint32_t size)
{
  uint32_t r = z % size;
  assert(r >= 0);
  assert(r < size);
  return r;
}

struct puk_hash_entry_s
{
  const void*key;
  void*data;
  struct puk_hash_entry_s*next;
};

struct puk_hashtable_s
{
  struct puk_hash_entry_s*table;
  int size;
  int load;
  puk_hash_hfunc_t hash_func;
  puk_hash_eqfunc_t eq_func;
};


void puk_hashtable_enumerator_init(puk_hashtable_enumerator_t e, puk_hashtable_t h)
{
  e->h = h;
  e->i = -1;
  e->current = NULL;
}
void puk_hashtable_enumerator_next2(puk_hashtable_enumerator_t e, void*key, void*data)
{
  if(e->current && e->current->next)
    {
      /* bump one bucket */
      e->current = e->current->next;
    }
  else
    {
      do
	{
	  e->i++;
	}
      while(e->i < 0 || (e->i < e->h->size && !e->h->table[e->i].key));
      e->current = e->h->table + e->i;
    }
  if(e->i < e->h->size)
    {
      if(key)
	*(const void**)key  = e->current->key;
      if(data)
	*(void**)data = e->current->data;
    }
  else
    {
      if(key)
	*(const void**)key  = NULL;
      if(data)
	*(void**)data = NULL;
    }
}

const static int puk_hash_primes[] = {
  11,17,23,29,37,47,59,71,89,107,131,163,197,239,293,353,431,521,631,761,919,
  1103,1327,1597,1931,2333,2801,3371,4049,4861,5839,7013,8419,10103,12143,14591,
  17519,21023,25229,30293,36353,43627,52361,62851,75431,90523, 108631, 130363,
  156437, 187751, 225307, 270371, 324449, 389357, 467237, 560689, 672827, 807403,
  968897, 1162687, 1395263, 1674319, 2009191, 2411033, 2893249, 3471899, 4166287,
  4999559, 5999471, 7199369
}; 

static void puk_hashtable_rehash(puk_hashtable_t h)
{
  int i = 0;
  int new_size;
  struct puk_hash_entry_s*new_table;
  while(h->size >= puk_hash_primes[i])
    {
      i++;
    }
  new_size = puk_hash_primes[i];
  assert(new_size > 0);
#ifdef DEBUG_HASH
  padico_print("Rehashing- size=%d; new_size=%d; load=%d\n", h->size, new_size, h->load);
#endif
  new_table = padico_calloc(new_size, sizeof(struct puk_hash_entry_s));
  for(i = 0; i < h->size; i++)
    {
      struct puk_hash_entry_s*source = h->table + i;
      int free_source = 0;
      while(source && source->key)
	{
	  uint32_t z = (*h->hash_func)(source->key);
	  struct puk_hash_entry_s*dest = new_table + puk_hash_to_index(z, new_size);
	  if(dest->key)
	    {
	      while(dest->next)
		{
		  dest = dest->next;
		}
	      assert(dest->next == NULL);
	      dest->next = padico_malloc(sizeof(struct puk_hash_entry_s));
	      dest = dest->next;
	    }
#ifdef DEBUG_HASH
	  padico_print("Rehashing- inserting key=%p; hash=%ud; index=%d\n",
		       source->key, z, puk_hash_to_index(z, new_size));
#endif
	  dest->key  = source->key;
	  dest->data = source->data;
	  dest->next = NULL;
	  if(free_source)
	    {
	      struct puk_hash_entry_s*n = source->next;
	      padico_free(source);
	      source = n;
	    }
	  else
	    {
	      source = source->next;
	      free_source = 1;
	    }
	}
    }
  padico_free(h->table);
  h->size  = new_size;
  h->table = new_table;
}


puk_hashtable_t puk_hashtable_new(puk_hash_hfunc_t hash_func,
						puk_hash_eqfunc_t eq_func)
{
  puk_hashtable_t h = padico_malloc(sizeof(struct puk_hashtable_s));
  h->size = puk_hashtable_init_size;
  h->load = 0;
  h->hash_func = hash_func;
  h->eq_func = eq_func;
  h->table = padico_calloc(h->size, sizeof(struct puk_hash_entry_s));
  return h;
}

void puk_hashtable_delete(puk_hashtable_t h)
{
  int i;
  for(i = 0; i < h->size; i++)
    {
      struct puk_hash_entry_s*e = h->table + i;
      e = e->next;
      while(e)
	{
	  struct puk_hash_entry_s*n = e->next;
	  padico_free(e);
	  e = n;
	}
    }
  padico_free(h->table);
  padico_free(h);
}

void puk_hashtable_insert(puk_hashtable_t h, const void*key, void*data)
{
  uint32_t z = (*h->hash_func)(key);
  struct puk_hash_entry_s*const e_init = h->table + puk_hash_to_index(z, h->size), *e = e_init;
  assert(key != NULL);
#ifdef DEBUG_HASH
  padico_print("Inserting- key=%p; hash=%ud; i=%d\n", key, z, puk_hash_to_index(z, h->size));
#endif
  while(e && e->key)
    {
      if((*h->eq_func)(key, e->key))
	{
	  e->data = data;
	  return;
	}
      e = e->next;
    }
  e = e_init;
  if(e->key)
    {
      while(e->next)
	{
	  e = e->next;
	}
      e->next = padico_malloc(sizeof(struct puk_hash_entry_s));
      e = e->next;
    }
  e->key  = key;
  e->data = data;
  e->next = NULL;
  h->load++;
  if(h->load >= h->size*3/4)
    puk_hashtable_rehash(h);
}

void* puk_hashtable_lookup(puk_hashtable_t h, const void*key)
{
  uint32_t z = (*h->hash_func)(key);
  struct puk_hash_entry_s*e = h->table + puk_hash_to_index(z, h->size);
  while(e && e->key)
    {
      if((*h->eq_func)(key, e->key))
	{
	  return e->data;
	}
      e = e->next;
    }
  return NULL;
}

void puk_hashtable_remove(puk_hashtable_t h, const void*key)
{
  uint32_t z = (*h->hash_func)(key);
  struct puk_hash_entry_s*e = h->table + puk_hash_to_index(z, h->size);
  struct puk_hash_entry_s*prev = NULL;
  while(e && e->key)
    {
      if((*h->eq_func)(key, e->key))
	{
	  if(e->next)
	    {
	      struct puk_hash_entry_s*next_e = e->next;
	      *e = *e->next;
	      padico_free(next_e);
	      prev = e;
	      e = e->next;
	    }
	  else
	    {
	      if(prev)
		{
		  padico_free(e);
		  e = NULL;
		}
	      else 
		{
		  e->key = NULL;
		  e->data = NULL;
		  e = NULL;
		}
	    }
	  h->load--;
	}
    }
}

int puk_hashtable_size(puk_hashtable_t h)
{
  return h->load;
}

/* *** hash functions collection *************************** */


uint32_t puk_hash_oneatatime(const unsigned char*key, int len)
{
  register uint32_t z = 0;
  register int i;
  for (i = 0; i < len; i++)
    {
      z += key[i];
      z += (z << 10);
      z ^= (z >> 6);
    }
  z += (z << 3);
  z ^= (z >> 11);
  z += (z << 15);
  return z;
}

int puk_hash_int_eq(const void*key1, const void*key2)
{
  int a = (long)key1;
  int b = (long)key2;
  return a == b;
}
uint32_t puk_hash_int(const void*key)
{
  int a = (long)key;
  return puk_hash_oneatatime((void*)&a, sizeof(int));
}

int puk_hash_string_eq(const void*key1, const void*key2)
{
  return !strcmp((const char*)key1, (const char*)key2);
}
uint32_t puk_hash_string(const void*key)
{
  const unsigned char*s;
  uint32_t z = 0;
  for(s = (const unsigned char*)key; *s; s++)
    {
      z += *s;
      z += (z << 10);
      z ^= (z >> 6);
    }
  z += (z << 3);
  z ^= (z >> 11);
  z += (z << 15);
  return z;
}

int puk_hash_pointer_eq(const void*key1, const void*key2)
{
  return key1 == key2;
}
uint32_t puk_hash_pointer(const void*key)
{
  const uint32_t z = (uint32_t)((((unsigned long)key) >> 4) & 0xFFFFFFFF);
  /*
  const unsigned char*s = (unsigned char*)&key;
  uint32_t z = 0;
  int i;
  for (i = 0; i < sizeof(void*); i++)
    {
      z += s[i];
      z += (z << 10);
      z ^= (z >> 6);
    }
  z += (z << 3);
  z ^= (z >> 11);
  z += (z << 15);
  */
  return z;
}

