/** @file
 * @brief Provides "pkg" module driver
 * @ingroup Puk
 */

#include "Puk-internals.h"
#include "Module.h"

PADICO_MODULE_HOOK(Puk);


struct puk_unit_pkg_s
{
  padico_mod_t sub_mod;
  padico_mod_t super_mod;
};
PUK_AUTO_TYPE(puk_unit_pkg);

static padico_rc_t puk_unit_pkg_load  (padico_unit_t unit);
static padico_rc_t puk_unit_pkg_start (padico_unit_t unit, puk_job_t job);
static padico_rc_t puk_unit_pkg_unload(padico_unit_t unit);
static padico_rc_t puk_unit_virtual_func (padico_unit_t unit);
static padico_rc_t puk_unit_virtual_start(padico_unit_t unit, puk_job_t job);


static const struct padico_loader_s puk_virtual_driver =
{
  "virtual",
  &puk_unit_virtual_func,
  &puk_unit_virtual_start,
  &puk_unit_virtual_func,
};

static const struct padico_loader_s puk_pkg_driver =
{
  "pkg",
  &puk_unit_pkg_load,
  &puk_unit_pkg_start,
  &puk_unit_pkg_unload,
};

void padico_puk_pkg_init()
{
  padico_puk_driver_add(&puk_pkg_driver);
  padico_puk_driver_add(&puk_virtual_driver);
}

/* *** driver: "virtual" *********************************** */

static padico_rc_t puk_unit_virtual_func(padico_unit_t unit)
{
  padico_warning("Virtual module with unit given!\n");
  return padico_rc_error("unit in virtual module\n");
}

static padico_rc_t puk_unit_virtual_start(padico_unit_t unit, puk_job_t job)
{
  padico_warning("Virtual module with unit given!\n");
  return padico_rc_error("unit in virtual module\n");
}


/* *** driver: "pkg" *************************************** */

static padico_rc_t puk_unit_pkg_load(padico_unit_t unit)
{
  puk_unit_pkg_t pkg;
  padico_rc_t rc = NULL;

  padico_trace("Puk: PKG LOAD **%s**\n", unit->name);
  if(!unit->driver_specific)
    {
      pkg = puk_unit_pkg_new();
      pkg->super_mod = unit->mod;
      pkg->sub_mod = puk_mod_getbyname(unit->name);
      if(pkg->sub_mod)
	{
	  rc = padico_rc_error("PKG load error: submodule %s already loaded", unit->name);
	  return rc;
	}
      rc = padico_puk_mod_open(&pkg->sub_mod, unit->name);
      if(padico_rc_iserror(rc))
	{
	  return rc;
	}
      pkg->sub_mod->inherits = unit->mod;
      unit->driver_specific = (void*)pkg;
    }
  else
    {
      pkg = (puk_unit_pkg_t)unit->driver_specific;
    }
  rc = padico_puk_mod_load(pkg->sub_mod);
  return rc;
}

static padico_rc_t puk_unit_pkg_start(padico_unit_t unit, puk_job_t job)
{
  puk_unit_pkg_t pkg = (puk_unit_pkg_t)unit->driver_specific;
  padico_rc_t rc = NULL;
  
  padico_trace("Puk: PKG RUN **%s**\n", unit->name);
  /* TODO: **warning** this works because a pkg is allowed to 
   *       have **only one** runnable unit.
   *       the sub_mod is started with the same job.
   */
  rc = padico_puk_mod_start(pkg->sub_mod, job);
  return rc;
}

static padico_rc_t puk_unit_pkg_unload(padico_unit_t unit)
{ 
  puk_unit_pkg_t pkg = (puk_unit_pkg_t)unit->driver_specific;
  padico_rc_t rc = NULL;

  padico_trace("Puk: PKG UNLOAD **%s**\n", unit->name);
  rc = padico_puk_mod_unload(pkg->sub_mod);
  pkg->sub_mod = NULL;
  puk_unit_pkg_delete(pkg);
  unit->driver_specific = NULL;
  return rc;
}

void puk_unit_pkg_ctor(puk_unit_pkg_t pkg)
{
  pkg->sub_mod = NULL;
  pkg->super_mod = NULL;
}

void puk_unit_pkg_dtor(puk_unit_pkg_t pkg) 
{
  if(pkg->sub_mod)
    padico_fatal("Puk: PKG internal error -- submods not unloaded\n");
}

