/** @file
 * @brief Automatic strings and various encoding functions
 * @ingroup Puk
 */

#include "Puk.h"
#include <string.h>
#include <stdio.h>
#include <stdarg.h>

void padico_string_grow(padico_string_t s, int size_inc)
{
  if(s->_size + size_inc >= s->_capacity)
    {
      s->_capacity *= 2;
      if(s->_size + size_inc >= s->_capacity)
	{
	  s->_capacity += size_inc;
	}
      s->_text = (char*)padico_realloc(s->_text, s->_capacity);
    }
}

void padico_string_printf(padico_string_t s, const char*fmt, ...)
{
  int rc;
  int restart;
  va_list ap;
  do {
    restart = 0;
    va_start(ap, fmt);
    rc = vsnprintf(s->_text, s->_capacity, fmt, ap);
    if((rc < 0) || (rc >= s->_capacity))
      {
	padico_string_grow(s, rc);
	restart = 1;
      }
    va_end(ap);
  }
  while(restart);
  s->_size = strlen(s->_text);
}

void padico_string_cat(padico_string_t s, const char*s2)
{
  padico_string_grow(s, strlen(s2));
  strcat(s->_text, s2);
  s->_size = strlen(s->_text);
}

void padico_string_catf(padico_string_t s, const char*fmt, ...)
{
  int rc;
  int restart;
  va_list ap;
  do {
    restart = 0;
    va_start(ap, fmt);
    rc = vsnprintf(s->_text+s->_size, s->_capacity - s->_size, fmt, ap);
    if((rc < 0) || (rc >= s->_capacity - s->_size))
      {
	restart = 1;
	padico_string_grow(s, rc);
      }
    va_end(ap);
  }
  while(restart);
  s->_size = strlen(s->_text);
}

/* ********************************************************* */
/* *** Encoding ******************************************** */

static const char*_hex_table =
"0123456789ABCDEF";

static const char _hex_reverse[] =
  {
    /*  32 */ -1, -1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, -1,
    /*  48 */  0,  1,  2,  3,   4,  5,  6,  7,   8,  9, -1, -1,  -1, -1, -1, -1,
    /*  64 */ -1, 10, 11, 12,  13, 14, 15, -1,  -1, -1, -1, -1,  -1, -1, -1, -1,
    /*  80 */ -1, -1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, -1,
    /*  96 */ -1, 10, 11, 12,  13, 14, 15, -1,  -1, -1, -1, -1,  -1, -1, -1, -1,
    /* 112 */ -1, -1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, -1
  };

/** @internal */
static inline char reverse_hex(char c)
{
  return ( (c > 32 && c < 127) ? _hex_reverse[c - 32] : -1);
}
static inline char encode_hex(unsigned char c)
{
  int i =(int)c;
  assert(i < 16);
  return _hex_table[i];
}
char*puk_hex_encode(const void*_bytes, int*len, void*(*alloc)(unsigned long))
{
  int i;
  const int out_len = *len*2+1;
  char*text = alloc?(*alloc)(out_len):padico_malloc(out_len);
  char*p = text;
  const unsigned char*bytes = _bytes;
  memset(text, 0, out_len);
  for(i = 0; i < *len; i++)
    {
      unsigned char c = bytes[i];
      *(p++) = encode_hex((unsigned char)((c & 0xF0) >> 4));
      *(p++) = encode_hex((unsigned char)(c & 0x0F));
    }
  *p = 0;
  *len = p - text;
  assert(*len == strlen(text));
  return text;
}

void*puk_hex_decode(const char*text, int*len, void*(*alloc)(unsigned long))
{
  const int out_len = ((*len + 1) / 2) + 1;
  char*bytes = alloc?(*alloc)(out_len):padico_malloc(out_len);
  char*p = bytes;
  assert(bytes != NULL && *len == strlen(text));
  while(*text)
    {
      *(p++) = (reverse_hex(*text) << 4) | reverse_hex(*(text+1));
      text += 2;
    }
  *len = p-bytes;
  return bytes;
}

char*puk_escape_encode(const void*bytes, int*_len, void*(*alloc)(unsigned long))
{
  int i;
  const int len = *_len;
  const int out_len = len * 3 + 2;
  char*text = alloc?(*alloc)(out_len):padico_malloc(out_len);
  char*p = text;
  memset(text, 0, out_len);
  for(i = 0; i < len; i++)
    {
      unsigned char c = ((unsigned char*)bytes)[i];
      if( ('a' <= c && c <= 'z') ||
	  ('A' <= c && c <= 'Z') ||
	  ('0' <= c && c <= '9'))
	{
	  *(p++) = c;
	}
      else
	{
	  *(p++) = '%';
	  *(p++) = _hex_table[(c & 0xF0) >> 4];
	  *(p++) = _hex_table[(c & 0x0F)];
	}
      *p = 0;
    }
  *_len = p - text;
  assert(*_len == strlen(text));
  return text;
}

void*puk_escape_decode(const char*text, int*len, void*(*alloc)(unsigned long))
{
  const int out_len = *len + 1;
  char*bytes = alloc?(*alloc)(out_len):padico_malloc(out_len);
  char*p = bytes;
  assert(bytes != NULL && *len == strlen(text));
  while(*text)
    {
      char c = *(text++);
      if(c == '%')
	{
	  *(p++) = (reverse_hex(*text) << 4) | reverse_hex(*(text+1));
	  text += 2;
	}
      else
	{
	  *(p++) = c;
	}
    }
  *len = p-bytes;
  return bytes;
}


/* *** base64 ********************************************** */

static const char*_base64 = 
"ABCDEFGHIJKLMNOPQRSTUVWXYZ"
"abcdefghijklmnopqrstuvwxyz"
"0123456789+/";

static const char _base64_reverse[] = 
  {
    /*  32 */ -1, -1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, 62,  -1, -1, -1, 63,
    /*  48 */ 52, 53, 54, 55,  56, 57, 58, 59,  60, 61, -1, -1,  -1, -1, -1, -1,
    /*  64 */ -1,  0,  1,  2,   3,  4,  5,  6,   7,  8,  9, 10,  11, 12, 13, 14,
    /*  80 */ 15, 16, 17, 18,  19, 20, 21, 22,  23, 24, 25, -1,  -1, -1, -1, -1,
    /*  96 */ -1, 26, 27, 28,  29, 30, 31, 32,  33, 34, 35, 36,  37, 38, 39, 40,
    /* 112 */ 41, 42, 43, 44,  45, 46, 47, 48,  49, 50, 51, -1,  -1, -1, -1, -1
  };
static inline char encode_one_base64(char c)
{
  return ( (c >= 0 && c < 64) ? _base64[(int)c] : -1);
}
static inline char decode_one_base64(char c)
{
  return ( (c > 32 && c < 127) ? _base64_reverse[c - 32] :  -1);
}

