/** @file
 * @brief Puk main functions
 * @ingroup Puk
 * @note Puk is not a true PadicoTM module. It is declared as
 * a module for compatibility with padico_module_self()
 */

#include <stdlib.h>
#include <setjmp.h>

#include "Module.h"
#include "Puk-internals.h"

PADICO_MODULE_HOOK(Puk); /* only a hook here. 
			  * For some obscure reasons, the real
			  * declaration is in puk-mod.c
			  */

/******************************************************/

static void puk_args_init(int argc, char* argv[]);

static struct
{
  char  hostname[256];
  int   argc;
  char**argv;
  const char*root;
} puk = { .root = NULL };

/******************************************************/

void padico_puk_bootstrap(int argc, char**argv)
{
  padico_puk_init(argc, argv);
  padico_puk_start();
  padico_puk_shutdown();
}

void padico_puk_init(int argc, char**argv)
{
  if(puk.root)
    return; /* Puk already initialized */
#ifdef PADICO_ENABLE_OUTPUT
  /* welcome banner */
  fprintf(stderr, "** Welcome to " PADICO_BUILD_STRING "\n");
#endif
  if(gethostname(puk.hostname, 256))
    padico_fatal("Puk: cannot get gethostname()\n");
  const char *_padico_root = padico_getenv("PADICO_ROOT");
  puk.root = _padico_root?:PADICO_ROOT;

  /* low-level init */
  puk_args_init(argc, argv);
  padico_puk_xml_init();        /* XML parser */
  padico_puk_mod_init();        /* module management */
  padico_puk_trace_init();      /* trace system */
  padico_puk_pkg_init();        /* driver: 'pkg' */
  padico_puk_lib_init();        /* routines for dynamic libraries */
  padico_puk_components_init(); /* component framework */
  padico_puk_monitor_init();    /* monitoring functions ('lsmod' and the likes) */
  padico_puk_boot_init();       /* driver: binary 'boot' */
}

void padico_puk_start(void)
{
  padico_mod_t startup_mod = NULL;
  padico_rc_t rc = NULL;
  /* load the startup module */
  const char* startup_name = padico_getattr("PADICO_STARTUP");
  if(startup_name == NULL)
    {
      padico_fatal("Puk: no startup module given.\n");
    }
  padico_print("Puk: loading startup module '%s'.\n", startup_name);
  rc = padico_puk_mod_open(&startup_mod, startup_name);
  if(padico_rc_iserror(rc))
    {
      padico_rc_show(rc);
      padico_fatal("Puk: startup module '%s' not found.\n", startup_name);
    }
  rc = padico_puk_mod_load(startup_mod);
  if(padico_rc_iserror(rc))
    {
      padico_rc_show(rc);
      padico_fatal("Puk: error while loading startup module.\n");
    }
  /* The startup mod is supposed to take the control of the process.
   * We never reach here.
   */
  padico_fatal("Puk: invalid startup module. It should never return from init!\n");
}

void padico_puk_shutdown(void)
{
  padico_puk_trace_finalize();
}

void puk_notimplemented()
{
  padico_fatal("Function not implemented\n");
}

/******************************************************/

int padico_getargc()
{ 
  return puk.argc; 
}

int* padico_getpargc()
{
  return (&puk.argc); 
}

char** padico_getargv()
{
  return puk.argv; 
}

char*** padico_getpargv() 
{
  return &puk.argv; 
}

void padico_setenv(const char *var, const char *val)
{
#ifdef HAVE_SETENV
  setenv(var, val, 1);
#else
  int l;
  char *p;
  l = strlen(var)+strlen(val)+2;
  p = (char*) padico_malloc(l);
  strcpy(p, var);
  strcat(p, "=");
  strcat(p, val);
  putenv(p);
#endif
}

const char*padico_getenv(const char*label)
{
  const char*value = getenv(label);
  return value;
}

const char*padico_hostname(void)
{
  return puk.hostname;
}

const char*padico_root(void)
{
  return puk.root;
}

/******************************************************/

static void puk_args_init(int argc, char* argv[])
{
  puk.argc = argc;
  puk.argv = padico_malloc((argc + 1) * sizeof(char*));
  memcpy(puk.argv, argv, (argc + 1) * sizeof(char*));
  /* build a fake argv[0] for Madeleine 2 */
  const char padico_d_name[] = "padico_d";
  char* padico_d = padico_malloc(strlen(padico_root()) + strlen(padico_d_name) + 2);
  strcpy(padico_d, padico_root());
  strcat(padico_d, "/");
  strcat(padico_d, padico_d_name);
  puk.argv[0] = padico_d;
}

